@@ -16,7 +16,7 @@
 # specific language governing permissions and limitations
 # under the License.
 
-__version__ = "5.2.0"
+__version__ = "5.3.0"
 __description__ = "Command Line Interface for Apache CloudStack"
 __maintainer__ = "The Apache CloudStack Team"
 __maintaineremail__ = "dev@cloudstack.apache.org"
@@ -48,7 +48,7 @@ config_fields = {'core': {}, 'ui': {}}
 
 # core
 config_fields['core']['asyncblock'] = 'true'
-config_fields['core']['paramcompletion'] = 'false'
+config_fields['core']['paramcompletion'] = 'true'
 config_fields['core']['cache_file'] = expanduser(config_dir + '/cache')
 config_fields['core']['history_file'] = expanduser(config_dir + '/history')
 config_fields['core']['log_file'] = expanduser(config_dir + '/log')
@@ -68,9 +68,12 @@ default_profile['username'] = 'admin'
 default_profile['password'] = 'password'
 default_profile['apikey'] = ''
 default_profile['secretkey'] = ''
+default_profile['verifysslcert'] = 'true'
+
 
 def write_config(get_attr, config_file):
-    global config_fields, mandatory_sections, default_profile, default_profile_name
+    global config_fields, mandatory_sections
+    global default_profile, default_profile_name
     config = ConfigParser()
     if os.path.exists(config_file):
         try:
@@ -128,7 +131,8 @@ def write_config(get_attr, config_file):
 
 
 def read_config(get_attr, set_attr, config_file):
-    global config_fields, config_dir, mandatory_sections, default_profile, default_profile_name
+    global config_fields, config_dir, mandatory_sections
+    global default_profile, default_profile_name
     if not os.path.exists(config_dir):
         os.makedirs(config_dir)
 
@@ -150,13 +154,20 @@ def read_config(get_attr, set_attr, config_file):
         print "After setting up, run the `sync` command to sync apis\n"
 
     missing_keys = []
-    profile = config.get('core', 'profile')
+    if config.has_option('core', 'profile'):
+        profile = config.get('core', 'profile')
+    else:
+        global default_profile_name
+        profile = default_profile_name
     if profile is None or profile == '' or profile in mandatory_sections:
         print "Server profile cannot be", profile
         sys.exit(1)
 
+    set_attr("profile_names", filter(lambda x: x != "core" and x != "ui",
+                                     config.sections()))
+
     if not config.has_section(profile):
-        print "Selected profile (%s) does not exit, will use the defaults" % profile
+        print "Selected profile (%s) does not exist, using defaults" % profile
         try:
             config.add_section(profile)
         except ValueError, e:
@@ -178,7 +189,7 @@ def read_config(get_attr, set_attr, config_file):
                     set_attr(key, config_fields[section][key])
                 else:
                     set_attr(key, default_profile[key])
-                missing_keys.append(key)
+                missing_keys.append("%s = %s" % (key, get_attr(key)))
             # Cosmetic fix for prompt
             if key == 'prompt':
                 set_attr(key, get_attr('prompt').strip() + " ")
